from functools import partial
from typing import List, Optional

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.collaboration_role import CollaboratorRole
from galileo_core.schemas.core.group_project import GroupProjectCollaboratorRequest, GroupProjectCollaboratorResponse


def share_project_with_group(
    project_id: UUID4,
    group_id: UUID4,
    role: CollaboratorRole = CollaboratorRole.viewer,
    config: Optional[GalileoConfig] = None,
) -> GroupProjectCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Sharing project {project_id} with group {group_id} with role {role}...")
    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.project_groups.format(project_id=project_id),
        json=[GroupProjectCollaboratorRequest(group_id=group_id, role=role).model_dump(mode="json")],
    )
    group_collaborators = [GroupProjectCollaboratorResponse.model_validate(group) for group in response_dict]
    logger.debug(f"Shared project {project_id} with group {group_id} with role {role}.")
    return group_collaborators[0]


def unshare_project_with_group(
    project_id: UUID4,
    group_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> None:
    config = config or GalileoConfig.get()
    logger.debug(f"Removing access for group {group_id} from project {project_id}")
    config.api_client.request(
        RequestMethod.DELETE, Routes.project_group.format(project_id=project_id, group_id=group_id)
    )
    logger.debug(f"Access for group {group_id} removed from project {project_id}.")


def list_group_project_collaborators(
    project_id: UUID4, config: Optional[GalileoConfig] = None
) -> List[GroupProjectCollaboratorResponse]:
    config = config or GalileoConfig.get()
    logger.debug("Listing groups the project has been shared with…")
    group_project_collaborators = paginated_request(
        partial(config.api_client.request, RequestMethod.GET, Routes.project_groups.format(project_id=project_id)),
        "collaborators",
    )
    groups = [GroupProjectCollaboratorResponse.model_validate(group) for group in group_project_collaborators]
    logger.debug(f"Listed all groups, found {len(groups)} groups.")
    return groups


def update_group_project_collaborator(
    project_id: UUID4,
    group_id: UUID4,
    role: CollaboratorRole = CollaboratorRole.viewer,
    config: Optional[GalileoConfig] = None,
) -> GroupProjectCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Updating project {project_id}: assigning role {role} to group {group_id}…")
    response_dict = config.api_client.request(
        RequestMethod.PATCH,
        Routes.project_group.format(project_id=project_id, group_id=group_id),
        json=GroupProjectCollaboratorRequest(group_id=group_id, role=role).model_dump(mode="json"),
    )
    group_project = GroupProjectCollaboratorResponse.model_validate(response_dict)
    logger.debug(f"Updated project {project_id}: assigned role {role} to group {group_id}…")
    return group_project



from typing import Any, Coroutine

from galileo_core.helpers.event_loop_thread_pool import EventLoopThreadPool

_thread_pool = EventLoopThreadPool(name="galileo_async_run")


def async_run(coroutine: Coroutine, wait_for_result: bool = True) -> Any:
    """
    Run an async coroutine synchronously.

    This function is useful for running async code in a synchronous context. This will
    ensure that the async code is run in a separate thread and the result is returned
    to the caller.

    Parameters
    ----------
    coroutine : Coroutine
        The coroutine to run.
    wait_for_result : bool, optional
        If True, the function will block until the coroutine is complete and return
        its result. If False, the function will return a concurrent.futures.Future
        object immediately.

    Returns
    -------
    Any
        The result of the coroutine.
    """
    return _thread_pool.submit(coroutine, wait_for_result=wait_for_result)



from datetime import datetime
from typing import Optional

from pydantic import UUID4, BaseModel

from galileo_core.schemas.core.collaboration_role import CollaboratorRole


class CreateApiKeyRequest(BaseModel):
    description: str
    expires_at: Optional[datetime] = None
    project_id: Optional[UUID4] = None
    project_role: Optional[CollaboratorRole] = None


class BaseApiKey(BaseModel):
    id: UUID4
    created_at: datetime
    updated_at: datetime
    last_used: Optional[datetime] = None
    truncated: str


class ApiKeyResponse(BaseApiKey, CreateApiKeyRequest):
    truncated: str


class CreateApiKeyResponse(ApiKeyResponse):
    api_key: str



from enum import Enum
from typing import List

from pydantic import BaseModel, Field

from galileo_core.schemas.shared.scorers.base_configs import (
    FinetunedScorerConfig,
    GeneratedScorerConfig,
    RegisteredScorerConfig,
)
from galileo_core.schemas.shared.scorers.scorers import GalileoScorer


class ScorerConfiguration(BaseModel):
    scorers: List[GalileoScorer] = Field(default_factory=list, description="List of Galileo scorers to enable.")
    registered_scorers: List[RegisteredScorerConfig] = Field(
        default_factory=list, description="List of registered scorers to enable."
    )
    generated_scorers: List[GeneratedScorerConfig] = Field(
        default_factory=list, description="List of generated scorers to enable."
    )
    finetuned_scorers: List[FinetunedScorerConfig] = Field(
        default_factory=list, description="List of finetuned scorers to enable."
    )


class ScorerInputType(str, Enum):
    """
    Enum for the type of input that a scorer can accept.
    """

    basic = "basic"  # Only the prompt input/output
    normalized = "normalized"  # Normalized trace/session



from abc import ABC
from logging import WARNING
from typing import Any, Dict, Optional

from galileo_core.helpers.logger import logger


class BaseGalileoException(Exception, ABC):
    """Base exception for all exceptions in galileo.

    Attributes:
        message: Human-readable error message
        error_code: Optional EMS catalog lookup key for standardized error handling
        LOG_LEVEL: Logging level for this exception type
    """

    LOG_LEVEL: int = WARNING

    def __init__(
        self,
        message: str,
        logging_extra: Optional[Dict[str, Any]] = None,
        *,
        error_code: Optional[int] = None,
    ) -> None:
        logger.log(self.LOG_LEVEL, message, extra=logging_extra)
        self.message: str = message
        self.error_code: Optional[int] = error_code
        super().__init__(self.message)



from enum import Enum
from typing import Any, Dict, Generator, List, Literal, Optional, Union

from pydantic import BaseModel, Field, RootModel, model_validator
from typing_extensions import Annotated


class MessageRole(str, Enum):
    agent = "agent"
    assistant = "assistant"
    developer = "developer"
    function = "function"
    system = "system"
    tool = "tool"
    user = "user"


class Message(BaseModel):
    content: str
    role: MessageRole
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List["ToolCall"]] = None

    @model_validator(mode="before")
    def _allow_null_content_with_tool_calling(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        # Some APIs (like OpenAI) often set content to None when there are tool calls.
        # This is a workaround to allow for that case without changing the type of the field.
        # TODO: Consider making the content field nullable.
        if data.get("content") is None:
            if data.get("tool_calls") is None:
                raise ValueError("at most one of 'content' and 'tool_calls' can be None, but both were None")
            # Deep copy and preserve key order
            data = {k: "" if k == "content" else v for k, v in data.items()}
        return data


class Messages(RootModel[List[Message]]):
    def __len__(self) -> int:
        return len(self.root)

    def __iter__(self) -> Generator[Message, None, None]:  # type: ignore[override]
        yield from self.root

    def __getitem__(self, item: int) -> Message:
        return self.root[item]


class ToolCall(BaseModel):
    id: str
    function: "ToolCallFunction"


class ToolCallFunction(BaseModel):
    name: str
    arguments: str


class EventType(str, Enum):
    """Types of events that can appear in reasoning/multi-turn model outputs."""

    message = "message"
    reasoning = "reasoning"
    internal_tool_call = "internal_tool_call"
    image_generation = "image_generation"
    mcp_call = "mcp_call"
    mcp_list_tools = "mcp_list_tools"
    mcp_approval_request = "mcp_approval_request"
    web_search_call = "web_search_call"


class EventStatus(str, Enum):
    """Common status values for events."""

    in_progress = "in_progress"
    completed = "completed"
    failed = "failed"
    cancelled = "cancelled"
    incomplete = "incomplete"


class BaseEvent(BaseModel):
    """Base class for all event types with common fields."""

    type: EventType = Field(description="The type of event")
    id: Optional[str] = Field(default=None, description="Unique identifier for the event")
    status: Optional[EventStatus] = Field(default=None, description="Status of the event")
    metadata: Optional[Dict[str, Any]] = Field(
        default=None, description="Provider-specific metadata and additional fields"
    )
    error_message: Optional[str] = Field(default=None, description="Error message if the event failed")


class MessageEvent(BaseEvent):
    """An output message from the model."""

    type: Literal[EventType.message] = EventType.message
    role: MessageRole = Field(description="Role of the message sender")
    content: Optional[str] = Field(default=None, description="Text content of the message")
    content_parts: Optional[List[Dict[str, Any]]] = Field(
        default=None, description="Structured content items (text, audio, images, etc.)"
    )


class ReasoningEvent(BaseEvent):
    """Internal reasoning/thinking from the model (e.g., OpenAI o1/o3 reasoning tokens)."""

    type: Literal[EventType.reasoning] = EventType.reasoning
    content: Optional[str] = Field(default=None, description="The reasoning/thinking content")
    summary: Optional[Union[str, List[Dict[str, Any]]]] = Field(default=None, description="Summary of the reasoning")


class InternalToolCall(BaseEvent):
    """A tool call executed internally by the model during reasoning.

    This represents internal tools like web search, code execution, file search, etc.
    that the model invokes (not user-defined functions or MCP tools).
    """

    type: Literal[EventType.internal_tool_call] = EventType.internal_tool_call
    name: str = Field(description="Name of the internal tool (e.g., 'web_search', 'code_interpreter', 'file_search')")
    input: Optional[Dict[str, Any]] = Field(default=None, description="Input/arguments to the tool call")
    output: Optional[Dict[str, Any]] = Field(default=None, description="Output/results from the tool call")


class WebSearchAction(BaseModel):
    """Action payload for a web search call event."""

    type: Literal["search"] = Field(description="Type of web search action")
    query: Optional[str] = Field(default=None, description="Search query string")
    sources: Optional[Any] = Field(default=None, description="Optional provider-specific sources")


class WebSearchCallEvent(BaseEvent):
    """An OpenAI-style web search call event."""

    type: Literal[EventType.web_search_call] = EventType.web_search_call
    action: WebSearchAction = Field(description="Web search action payload")


class ImageGenerationEvent(BaseEvent):
    """An image generation event from the model."""

    type: Literal[EventType.image_generation] = EventType.image_generation
    prompt: Optional[str] = Field(default=None, description="The prompt used for image generation")
    images: Optional[List[Dict[str, Any]]] = Field(
        default=None, description="Generated images with URLs or base64 data"
    )
    model: Optional[str] = Field(default=None, description="Image generation model used")


class MCPCallEvent(BaseEvent):
    """A Model Context Protocol (MCP) tool call.

    MCP is a protocol for connecting LLMs to external tools/data sources.
    This is distinct from internal tools because it involves external integrations.
    """

    type: Literal[EventType.mcp_call] = EventType.mcp_call
    tool_name: Optional[str] = Field(default=None, description="Name of the MCP tool being called")
    server_name: Optional[str] = Field(default=None, description="Name of the MCP server")
    arguments: Optional[Dict[str, Any]] = Field(default=None, description="Arguments for the MCP tool call")
    result: Optional[Dict[str, Any]] = Field(default=None, description="Result from the MCP tool call")


class MCPListToolsEvent(BaseEvent):
    """MCP list tools event - when the model queries available MCP tools."""

    type: Literal[EventType.mcp_list_tools] = EventType.mcp_list_tools
    server_name: Optional[str] = Field(default=None, description="Name of the MCP server")
    tools: Optional[List[Dict[str, Any]]] = Field(default=None, description="List of available MCP tools")


class MCPApprovalRequestEvent(BaseEvent):
    """MCP approval request - when human approval is needed for an MCP tool call."""

    type: Literal[EventType.mcp_approval_request] = EventType.mcp_approval_request
    tool_name: Optional[str] = Field(default=None, description="Name of the MCP tool requiring approval")
    tool_invocation: Optional[Dict[str, Any]] = Field(
        default=None, description="Details of the tool invocation requiring approval"
    )
    approved: Optional[bool] = Field(default=None, description="Whether the request was approved")


# Union of all event types with discriminator
Event = Annotated[
    Union[
        MessageEvent,
        ReasoningEvent,
        InternalToolCall,
        WebSearchCallEvent,
        ImageGenerationEvent,
        MCPCallEvent,
        MCPListToolsEvent,
        MCPApprovalRequestEvent,
    ],
    Field(discriminator="type"),
]



from datetime import datetime, timezone
from typing import Dict, List, Optional, Sequence, Type, TypeAlias, Union

from pydantic import UUID4, BaseModel, Field, TypeAdapter, field_validator
from pydantic_partial import PartialModelMixin
from typing_extensions import Annotated, Any, get_args

from galileo_core.schemas.logging.session import BaseSession
from galileo_core.schemas.logging.span import BaseAgentSpan, BaseRetrieverSpan, BaseToolSpan, BaseWorkflowSpan, LlmSpan
from galileo_core.schemas.logging.step import BaseStep, StepAllowedInputType
from galileo_core.schemas.logging.trace import BaseTrace
from galileo_core.schemas.shared.feedback import FeedbackRatingInfo


class RecordIdsWithMetrics(BaseModel):
    id: UUID4
    project_id: UUID4
    run_id: UUID4
    created_at: datetime
    updated_at: Optional[datetime] = Field(default=None)
    metrics: Dict[str, Any]


class BaseRecord(PartialModelMixin, BaseStep):
    id: UUID4 = Field(title="ID", description="Galileo ID of the session, trace or span")
    session_id: UUID4 = Field(title="Session ID", description="Galileo ID of the session")
    trace_id: Optional[UUID4] = Field(
        default=None,
        title="Trace ID",
        description="Galileo ID of the trace containing the span (or the same value as id for a trace)",
    )
    project_id: UUID4 = Field(
        title="Project ID", description="Galileo ID of the project associated with this trace or span"
    )
    run_id: UUID4 = Field(
        title="Run ID",
        description="Galileo ID of the run (log stream or experiment) associated with this trace or span",
    )
    updated_at: Optional[datetime] = Field(
        default=None, title="Last Updated", description="Timestamp of the session or trace or span's last update"
    )
    has_children: Optional[bool] = Field(default=None, description="Whether or not this trace or span has child spans")
    metrics_batch_id: Optional[UUID4] = Field(
        default=None, description="Galileo ID of the metrics batch associated with this trace or span"
    )
    session_batch_id: Optional[UUID4] = Field(
        default=None, description="Galileo ID of the metrics batch associated with this trace or span"
    )
    feedback_rating_info: Dict[str, FeedbackRatingInfo] = Field(
        default_factory=dict, description="Feedback information related to the record"
    )

    def to_record_ids_with_metrics(self) -> RecordIdsWithMetrics:
        return RecordIdsWithMetrics(
            id=self.id,
            project_id=self.project_id,
            run_id=self.run_id,
            created_at=self.created_at,
            updated_at=self.updated_at,
            metrics=self.metrics.model_dump(exclude_unset=True, exclude_none=True),
        )

    @field_validator("created_at", "updated_at", mode="after")
    def ensure_tz_aware(cls, value: Optional[datetime]) -> Optional[datetime]:
        if value is None:
            return None
        if value.tzinfo is None:
            value = value.replace(tzinfo=timezone.utc)
        return value


class TraceRecord(BaseTrace, BaseRecord):
    session_id: UUID4 = Field(
        title="Session ID",
        description="Galileo ID of the session containing the trace (or the same value as id for a trace)",
    )
    trace_id: UUID4 = Field(
        title="Trace ID",
        description="Galileo ID of the trace containing the span (or the same value as id for a trace)",
    )
    id: UUID4 = Field(title="ID", description="Galileo ID of the trace")
    is_complete: bool = Field(default=True, description="Whether the trace is complete or not")


class SessionRecord(BaseSession, BaseRecord):
    id: UUID4 = Field(title="ID", description="Galileo ID of the session")
    input: StepAllowedInputType = Field(default="", validate_default=True)

    @field_validator("input", mode="after")
    def ensure_non_null_input(cls, value: Optional[StepAllowedInputType]) -> StepAllowedInputType:
        if value is None:
            value = ""
        return value


class BaseSpanRecord(BaseRecord):
    trace_id: UUID4 = Field(
        title="Trace ID",
        description="Galileo ID of the trace containing the span (or the same value as id for a trace)",
    )
    parent_id: UUID4 = Field(title="Parent ID", description="Galileo ID of the parent of this span")
    is_complete: bool = Field(default=True, description="Whether the parent trace is complete or not")


class WorkflowSpanRecord(BaseWorkflowSpan, BaseSpanRecord):
    session_id: UUID4 = Field(
        title="Session ID",
        description="Galileo ID of the session containing the trace (or the same value as id for a trace)",
    )
    parent_id: UUID4 = Field(title="Parent ID", description="Galileo ID of the parent of this span")
    id: UUID4 = Field(title="ID", description="Galileo ID of the session, trace or span")


class AgentSpanRecord(BaseAgentSpan, BaseSpanRecord):
    session_id: UUID4 = Field(
        title="Session ID",
        description="Galileo ID of the session containing the trace (or the same value as id for a trace)",
    )
    parent_id: UUID4 = Field(title="Parent ID", description="Galileo ID of the parent of this span")
    id: UUID4 = Field(title="ID", description="Galileo ID of the session, trace or span")


class LlmSpanRecord(LlmSpan, BaseSpanRecord):
    session_id: UUID4 = Field(
        title="Session ID",
        description="Galileo ID of the session containing the trace (or the same value as id for a trace)",
    )
    parent_id: UUID4 = Field(title="Parent ID", description="Galileo ID of the parent of this span")
    id: UUID4 = Field(title="ID", description="Galileo ID of the session, trace or span")


class ToolSpanRecord(BaseToolSpan, BaseSpanRecord):
    session_id: UUID4 = Field(
        title="Session ID",
        description="Galileo ID of the session containing the trace (or the same value as id for a trace)",
    )
    parent_id: UUID4 = Field(title="Parent ID", description="Galileo ID of the parent of this span")
    id: UUID4 = Field(title="ID", description="Galileo ID of the session, trace or span")


class RetrieverSpanRecord(BaseRetrieverSpan, BaseSpanRecord):
    session_id: UUID4 = Field(
        title="Session ID",
        description="Galileo ID of the session containing the trace (or the same value as id for a trace)",
    )
    parent_id: UUID4 = Field(title="Parent ID", description="Galileo ID of the parent of this span")
    id: UUID4 = Field(title="ID", description="Galileo ID of the session, trace or span")


SpanRecord = Annotated[
    Union[
        AgentSpanRecord,
        WorkflowSpanRecord,
        LlmSpanRecord,
        ToolSpanRecord,
        RetrieverSpanRecord,
    ],
    Field(discriminator="type"),
]


SpanRecordAdapter: TypeAdapter[SpanRecord] = TypeAdapter(SpanRecord)


SpanRecordTypes: List[Type[BaseRecord]] = list(get_args(get_args(SpanRecord)[0]))


TraceOrSpanRecord = Annotated[
    Union[
        TraceRecord,
        SpanRecord,
    ],
    Field(discriminator="type"),
]


TraceOrSpanRecordAdapter: TypeAdapter[TraceOrSpanRecord] = TypeAdapter(TraceOrSpanRecord)


RecordType = Annotated[
    Union[
        TraceRecord,
        SpanRecord,
        SessionRecord,
    ],
    Field(discriminator="type"),
]


RecordTypeAdapter: TypeAdapter[RecordType] = TypeAdapter(RecordType)


class RecordWithChildSpans(BaseModel):
    spans: Sequence["SpanRecordWithChildren"] = Field(default_factory=list)


class TraceRecordWithChildren(TraceRecord, RecordWithChildSpans):
    pass


class RecordWithChildTraces(BaseModel):
    traces: Sequence["TraceRecordWithChildren"] = Field(default_factory=list)


class WorkflowSpanRecordWithChildren(WorkflowSpanRecord, RecordWithChildSpans):
    pass


class AgentSpanRecordWithChildren(AgentSpanRecord, RecordWithChildSpans):
    pass


class RetrieverSpanRecordWithChildren(RetrieverSpanRecord, RecordWithChildSpans):
    pass


class ToolSpanRecordWithChildren(ToolSpanRecord, RecordWithChildSpans):
    pass


SpanRecordWithChildren = Annotated[
    Union[
        AgentSpanRecordWithChildren,
        WorkflowSpanRecordWithChildren,
        LlmSpanRecord,
        ToolSpanRecordWithChildren,
        RetrieverSpanRecordWithChildren,
    ],
    Field(discriminator="type"),
]


class SessionRecordWithChildren(SessionRecord, RecordWithChildTraces):
    pass


SpanRecordWithChildrenAdapter: TypeAdapter[SpanRecordWithChildren] = TypeAdapter(SpanRecordWithChildren)


PartialAgentSpanRecord: TypeAlias = AgentSpanRecord.model_as_partial()  # type: ignore[valid-type]
PartialWorkflowSpanRecord: TypeAlias = WorkflowSpanRecord.model_as_partial()  # type: ignore[valid-type]
PartialLlmSpanRecord: TypeAlias = LlmSpanRecord.model_as_partial()  # type: ignore[valid-type]
PartialToolSpanRecord: TypeAlias = ToolSpanRecord.model_as_partial()  # type: ignore[valid-type]
PartialRetrieverSpanRecord: TypeAlias = RetrieverSpanRecord.model_as_partial()  # type: ignore[valid-type]
PartialTraceRecord: TypeAlias = TraceRecord.model_as_partial()  # type: ignore[valid-type]
PartialSessionRecord: TypeAlias = SessionRecord.model_as_partial()  # type: ignore[valid-type]


PartialSpanRecord: TypeAlias = Annotated[
    Union[
        PartialAgentSpanRecord,
        PartialWorkflowSpanRecord,
        PartialLlmSpanRecord,
        PartialToolSpanRecord,
        PartialRetrieverSpanRecord,
    ],
    Field(discriminator="type"),
]


PartialSpanRecordAdapter: TypeAdapter[PartialSpanRecord] = TypeAdapter(PartialSpanRecord)  # type: ignore[valid-type]


PartialSpanRecordTypes: List[Type[BaseRecord]] = list(get_args(get_args(PartialSpanRecord)[0]))


PartialRecordType: TypeAlias = Annotated[
    Union[
        PartialTraceRecord,
        PartialSpanRecord,
        PartialSessionRecord,
    ],
    Field(discriminator="type"),
]


PartialRecordTypeAdapter: TypeAdapter[PartialRecordType] = TypeAdapter(PartialRecordType)  # type: ignore[valid-type]



from datetime import timedelta
from functools import cached_property
from os import getenv
from typing import Dict, List, Optional, Set
from uuid import UUID

from pydantic import UUID4, ConfigDict, Field, field_validator, model_validator

from galileo_core.schemas.protect.payload import Payload
from galileo_core.schemas.protect.rule import Rule
from galileo_core.schemas.protect.ruleset import RulesetsMixin


class Request(RulesetsMixin):
    payload: Payload = Field(description="Payload to be processed.")
    project_name: Optional[str] = Field(default=None, description="Project name.", validate_default=True)
    project_id: Optional[UUID4] = Field(default=None, description="Project ID.", validate_default=True)
    stage_name: Optional[str] = Field(default=None, description="Stage name.", validate_default=True)
    stage_id: Optional[UUID4] = Field(default=None, description="Stage ID.", validate_default=True)
    stage_version: Optional[int] = Field(
        default=None,
        description="Stage version to use for the request, if it's a central stage with a previously registered version.",
        validate_default=True,
    )
    timeout: float = Field(
        default=timedelta(minutes=5).total_seconds(),
        description="Optional timeout for the guardrail execution in seconds. This is not the timeout for the request. If not set, a default timeout of 5 minutes will be used.",
    )
    metadata: Optional[Dict[str, str]] = Field(
        default=None,
        description="Optional additional metadata. This will be echoed back in the response.",
    )
    headers: Optional[Dict[str, str]] = Field(
        default=None,
        description="Optional additional HTTP headers that should be included in the response.",
    )

    model_config = ConfigDict(populate_by_name=True)

    @field_validator("project_name", mode="before")
    def project_name_from_env(cls, value: Optional[str]) -> Optional[str]:
        env_value = getenv("GALILEO_PROJECT_NAME", getenv("GALILEO_PROJECT"))
        if value is None and env_value:
            value = env_value
        return value

    @field_validator("project_id", mode="before")
    def project_id_from_env(cls, value: Optional[UUID4]) -> Optional[UUID4]:
        env_value = getenv("GALILEO_PROJECT_ID")
        if value is None and env_value:
            value = UUID(env_value)
        return value

    @field_validator("stage_name", mode="before")
    def stage_name_from_env(cls, value: Optional[str]) -> Optional[str]:
        env_value = getenv("GALILEO_STAGE_NAME")
        if value is None and env_value:
            value = env_value
        return value

    @field_validator("stage_id", mode="before")
    def stage_id_from_env(cls, value: Optional[UUID4]) -> Optional[UUID4]:
        env_value = getenv("GALILEO_STAGE_ID")
        if value is None and env_value:
            value = UUID(env_value)
        return value

    @field_validator("stage_version", mode="before")
    def stage_version_from_env(cls, value: Optional[int]) -> Optional[int]:
        env_value = getenv("GALILEO_STAGE_VERSION")
        if value is None and env_value:
            value = int(env_value)
        return value

    @model_validator(mode="after")
    def validate_project_stage(self) -> "Request":
        """
        Validate that one of:
        1. stage_id
        2. stage_name and project_id
        3. stage_name and project_name
        are provided.

        Returns
        -------
        Request
            The request object.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if (self.stage_id) or (self.stage_name and self.project_id) or (self.stage_name and self.project_name):
            return self
        else:
            raise ValueError(
                "Either stage_id or stage_name and project_id or stage_name and project_name must be provided."
            )

    @cached_property
    def rules(self) -> List[Rule]:
        rules: List[Rule] = []
        for ruleset in self.rulesets:
            rules.extend(ruleset.rules)
        return rules

    @cached_property
    def metrics(self) -> Set[str]:
        metrics_to_compute = []
        for rule in self.rules:
            metrics_to_compute.append(rule.metric)
        return set(metrics_to_compute)

    @cached_property
    def timeout_ns(self) -> float:
        return self.timeout * 1e9



"""
Scorer utility functions and constants.

This module contains utility functions for working with scorer names and metric names,
including mappings between metric names and their corresponding scorer names.
"""

from re import match
from typing import Optional

from galileo_core.schemas.shared.scorers.scorer_name import ScorerName

# Used to validate custom scorer names.
SCORER_NAME_REGEX = r"^[\w -]+$"


METRIC_TO_SCORER_NAME = {
    # The luna/plus mess.
    "groundedness": ScorerName.context_adherence,
    "completeness_gpt": ScorerName.completeness,
    "retriever_attribution": ScorerName.chunk_attribution_utilization,
    "retriever_utilization": ScorerName.chunk_attribution_utilization,
    "rag_nli_adherence": ScorerName.context_adherence_luna,
    "context_adherence_luna": ScorerName.context_adherence_luna,
    "rag_nli_completeness": ScorerName.completeness_luna,
    "rag_nli_retriever_attribution": ScorerName.chunk_attribution_utilization_luna,
    "rag_nli_retriever_utilization": ScorerName.chunk_attribution_utilization_luna,
    "rag_nli_retriever_relevance": ScorerName.chunk_attribution_utilization_luna,
    # The output scorers.
    "pii": ScorerName.output_pii,
    "sexist": ScorerName.output_sexist_luna,
    "tone": ScorerName.output_tone,
    "toxicity": ScorerName.output_toxicity_luna,
    "input_toxicity": ScorerName.input_toxicity_luna,
    "input_sexist": ScorerName.input_sexist_luna,
    # Others.
    "factuality": ScorerName.correctness,
    "prompt_injection_gpt": ScorerName.prompt_injection,
    "prompt_injection": ScorerName.prompt_injection_luna,
    "sexist_gpt": ScorerName.output_sexist,
    "input_sexist_gpt": ScorerName.input_sexist,
    "toxicity_gpt": ScorerName.output_toxicity,
    "input_toxicity_gpt": ScorerName.input_toxicity,
}


# Reverse mapping from scorer names to all associated metric names.
SCORER_TO_ALL_METRICS = {}
for scorer in ScorerName:
    # Find all metrics that map to this scorer
    mapped_metrics = [
        metric_name for metric_name, scorer_name in METRIC_TO_SCORER_NAME.items() if scorer_name == scorer
    ]

    # If no metrics map to this scorer, use the scorer's own value as the metric
    if not mapped_metrics:
        mapped_metrics = [scorer.value]

    SCORER_TO_ALL_METRICS[scorer.value] = mapped_metrics


def check_scorer_name(name: str) -> str:
    """
    Check if name contains only letters, numbers, space, - and _.

    Parameters
    ----------
    name : str
        The scorer name to validate.

    Returns
    -------
    str
        The validated scorer name.

    Raises
    ------
    ValueError
        If the scorer name contains invalid characters.
    """
    if not bool(match(SCORER_NAME_REGEX, name)):
        raise ValueError("Scorer name cannot contain special characters, only letters, numbers, space, - and _.")
    return name


def get_scorer_name_from_metric_name(metric_name: str) -> Optional[ScorerName]:
    """
    Get the scorer name from the metric name and return None if no scorer matches.

    This function first checks the METRIC_TO_SCORER_NAME mapping for special cases
    where the metric name doesn't directly correspond to the scorer name. If not found,
    it tries to match the metric name directly to a ScorerName enum value.

    Parameters
    ----------
    metric_name : str
        The metric name to look up. Leading underscores will be stripped.

    Returns
    -------
    Optional[ScorerName]
        The corresponding ScorerName enum value, or None if no match is found.
    """
    metric_name = metric_name.lstrip("_")

    # For metrics that don't match their scorer name.
    scorer_name = METRIC_TO_SCORER_NAME.get(metric_name)
    if scorer_name is None:
        try:
            # For metrics that do match their scorer name.
            scorer_name = ScorerName(metric_name)
        except ValueError:
            # Return None if metric name not in ScorerName enum.
            return None
    return scorer_name



from typing import Union

from pydantic import Field
from typing_extensions import Annotated

from galileo_core.schemas.core.integration.anthropic import AnthropicIntegrationCreate
from galileo_core.schemas.core.integration.aws_bedrock import AwsBedrockIntegrationCreate
from galileo_core.schemas.core.integration.azure import AzureIntegrationCreate
from galileo_core.schemas.core.integration.mistral import MistralIntegrationCreate
from galileo_core.schemas.core.integration.open_ai import OpenAIIntegrationCreate
from galileo_core.schemas.core.integration.vertex_ai import VertexAIIntegrationCreate
from galileo_core.schemas.core.integration.writer import WriterIntegrationCreate

IntegrationUnion = Union[
    AnthropicIntegrationCreate,
    MistralIntegrationCreate,
    OpenAIIntegrationCreate,
    VertexAIIntegrationCreate,
    WriterIntegrationCreate,
    AzureIntegrationCreate,
    AwsBedrockIntegrationCreate,
]

CreateIntegrationModel = Annotated[IntegrationUnion, Field(discriminator="name")]



from enum import Enum


class SystemRole(str, Enum):
    # These are prefixed with 'system_' to differentiate from the org-level roles in Authz
    system_admin = "system_admin"
    system_user = "system_user"


class UserRole(str, Enum):
    admin = "admin"
    manager = "manager"
    user = "user"
    read_only = "read_only"
